@@ -4,17 +4,17 @@
     SPDX-License-Identifier: LGPL-2.1-only OR LGPL-3.0-only OR LicenseRef-KDE-Accepted-LGPL
 */
 #include "shm_pool.h"
-#include "event_queue.h"
 #include "buffer_p.h"
+#include "event_queue.h"
 #include "logging.h"
 #include "wayland_pointer_p.h"
 // Qt
 #include <QDebug>
 #include <QImage>
-#include <QTemporaryFile>
 // system
-#include <unistd.h>
+#include <fcntl.h>
 #include <sys/mman.h>
+#include <unistd.h>
 // wayland
 #include <wayland-client-protocol.h>
 
@@ -22,7 +22,6 @@ namespace KWayland
 {
 namespace Client
 {
-
 class Q_DECL_HIDDEN ShmPool::Private
 {
 public:
@@ -33,23 +32,22 @@ public:
     WaylandPointer<wl_shm, wl_shm_destroy> shm;
     WaylandPointer<wl_shm_pool, wl_shm_pool_destroy> pool;
     void *poolData = nullptr;
+    int fd = -1;
     int32_t size = 1024;
-    QScopedPointer<QTemporaryFile> tmpFile;
     bool valid = false;
     int offset = 0;
     QList<QSharedPointer<Buffer>> buffers;
     EventQueue *queue = nullptr;
+
 private:
     ShmPool *q;
 };
 
 ShmPool::Private::Private(ShmPool *q)
-    : tmpFile(new QTemporaryFile())
-    , q(q)
+    : q(q)
 {
 }
 
-
 ShmPool::ShmPool(QObject *parent)
     : QObject(parent)
     , d(new Private(this))
@@ -68,9 +66,12 @@ void ShmPool::release()
         munmap(d->poolData, d->size);
         d->poolData = nullptr;
     }
+    if (d->fd != -1) {
+        close(d->fd);
+        d->fd = -1;
+    }
     d->pool.release();
     d->shm.release();
-    d->tmpFile->close();
     d->valid = false;
     d->offset = 0;
 }
@@ -85,9 +86,12 @@ void ShmPool::destroy()
         munmap(d->poolData, d->size);
         d->poolData = nullptr;
     }
+    if (d->fd != -1) {
+        close(d->fd);
+        d->fd = -1;
+    }
     d->pool.destroy();
     d->shm.destroy();
-    d->tmpFile->close();
     d->valid = false;
     d->offset = 0;
 }
@@ -112,21 +116,39 @@ EventQueue *ShmPool::eventQueue()
 
 bool ShmPool::Private::createPool()
 {
-    if (!tmpFile->open()) {
+#if HAVE_MEMFD
+    fd = memfd_create("kwayland-shared", MFD_CLOEXEC | MFD_ALLOW_SEALING);
+    if (fd >= 0) {
+        fcntl(fd, F_ADD_SEALS, F_SEAL_SHRINK | F_SEAL_SEAL);
+    } else
+#endif
+    {
+        char templateName[] = "/tmp/kwayland-shared-XXXXXX";
+        fd = mkstemp(templateName);
+        if (fd >= 0) {
+            unlink(templateName);
+
+            int flags = fcntl(fd, F_GETFD);
+            if (flags == -1 || fcntl(fd, F_SETFD, flags | FD_CLOEXEC) == -1) {
+                close(fd);
+                fd = -1;
+            }
+        }
+    }
+
+    if (fd == -1) {
         qCDebug(KWAYLAND_CLIENT) << "Could not open temporary file for Shm pool";
         return false;
     }
-    if (unlink(tmpFile->fileName().toUtf8().constData()) != 0) {
-        qCDebug(KWAYLAND_CLIENT) << "Unlinking temporary file for Shm pool from file system failed";
-    }
-    if (ftruncate(tmpFile->handle(), size) < 0) {
+
+    if (ftruncate(fd, size) < 0) {
         qCDebug(KWAYLAND_CLIENT) << "Could not set size for Shm pool file";
         return false;
     }
-    poolData = mmap(nullptr, size, PROT_READ | PROT_WRITE, MAP_SHARED, tmpFile->handle(), 0);
-    pool.setup(wl_shm_create_pool(shm, tmpFile->handle(), size));
+    poolData = mmap(nullptr, size, PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);
+    pool.setup(wl_shm_create_pool(shm, fd, size));
 
-    if (!poolData || !pool) {
+    if (poolData == MAP_FAILED || !pool) {
         qCDebug(KWAYLAND_CLIENT) << "Creating Shm pool failed";
         return false;
     }
@@ -135,15 +157,15 @@ bool ShmPool::Private::createPool()
 
 bool ShmPool::Private::resizePool(int32_t newSize)
 {
-    if (ftruncate(tmpFile->handle(), newSize) < 0) {
+    if (ftruncate(fd, newSize) < 0) {
         qCDebug(KWAYLAND_CLIENT) << "Could not set new size for Shm pool file";
         return false;
     }
     wl_shm_pool_resize(pool, newSize);
     munmap(poolData, size);
-    poolData = mmap(nullptr, newSize, PROT_READ | PROT_WRITE, MAP_SHARED, tmpFile->handle(), 0);
+    poolData = mmap(nullptr, newSize, PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);
     size = newSize;
-    if (!poolData) {
+    if (poolData == MAP_FAILED) {
         qCDebug(KWAYLAND_CLIENT) << "Resizing Shm pool failed";
         return false;
     }
@@ -151,7 +173,8 @@ bool ShmPool::Private::resizePool(int32_t newSize)
     return true;
 }
 
-namespace {
+namespace
+{
 static Buffer::Format toBufferFormat(const QImage &image)
 {
     switch (image.format()) {
@@ -169,7 +192,7 @@ static Buffer::Format toBufferFormat(const QImage &image)
 }
 }
 
-Buffer::Ptr ShmPool::createBuffer(const QImage& image)
+Buffer::Ptr ShmPool::createBuffer(const QImage &image)
 {
     if (image.isNull() || !d->valid) {
         return QWeakPointer<Buffer>();
@@ -201,7 +224,8 @@ Buffer::Ptr ShmPool::createBuffer(const QSize &size, int32_t stride, const void
     return QWeakPointer<Buffer>(*it);
 }
 
-namespace {
+namespace
+{
 static wl_shm_format toWaylandFormat(Buffer::Format format)
 {
     switch (format) {
@@ -243,8 +267,7 @@ QList<QSharedPointer<Buffer>>::iterator ShmPool::Private::getBuffer(const QSize
         }
     }
     // we don't have a buffer which we could reuse - need to create a new one
-    wl_buffer *native = wl_shm_pool_create_buffer(pool, offset, s.width(), s.height(),
-                                                  stride, toWaylandFormat(format));
+    wl_buffer *native = wl_shm_pool_create_buffer(pool, offset, s.width(), s.height(), stride, toWaylandFormat(format));
     if (!native) {
         return buffers.end();
     }
@@ -262,7 +285,7 @@ bool ShmPool::isValid() const
     return d->valid;
 }
 
-void* ShmPool::poolAddress() const
+void *ShmPool::poolAddress() const
 {
     return d->poolData;
 }
